@@ -9,7 +9,7 @@ require Exporter;
 # Items to export into callers namespace by default. Note: do not export
 # names by default without a very good reason. Use EXPORT_OK instead.
 # Do not simply export all your public functions/methods/constants.
-$VERSION = '0.91';
+$VERSION = '0.95';
 
 
 # Preloaded methods go here.
@@ -20,9 +20,9 @@ use File::stat;
 use Carp;
 use Time::HiRes qw ( time sleep ); #import hires microsecond timers
 
-sub SEEK_SET {0;}
-sub SEEK_START {1;}
-sub SEEK_END {2;}
+sub SEEK_SET   () {0;}
+sub SEEK_CUR () {1;}
+sub SEEK_END   () {2;}
 
 
 sub interval {
@@ -38,9 +38,12 @@ sub interval {
 sub logit {
     my $object=shift;
     my @call=caller(1);
-    print STDERR 
+    print # STDERR 
 #	time()." ".
-	    $call[3]." ".$object->{"input"}." ".join("",@_)."\n"
+	"\033[7m".
+	    $call[3]." ".$object->{"input"}." ".join("",@_).
+		"\033[0m".
+		    "\n"
 	if $object->debug;
 }
 
@@ -208,7 +211,7 @@ sub DESTROY {
   my($this) = shift(@_);
   close($this->{"handle"});
   undef $this;
-  return undef;
+  return;
 }
 
 sub CLOSE {
@@ -227,7 +230,7 @@ sub new {
     } else {
 	if (($#_ % 2) != 1) {
 	    croak "Odd number of parameters for new";
-	    return undef;
+	    return;
 	}
 	%params=@_;
     }
@@ -249,6 +252,7 @@ sub new {
 			 ($object->maxinterval*$object->adjustafter));
     $object->{"debug"}=($params{'debug'} || 0);
     $object->{"tail"}=($params{'tail'} || 0);
+    $object->{"maxbuf"}=($params{'maxbuf'} || 16384);
     if (defined $params{'reset_tail'}) {
         $object->{"reset_tail"} = $params{'reset_tail'};
     } else {
@@ -258,6 +262,8 @@ sub new {
     $object->{"lastread"}=0;
     $object->{"sleepcount"}=0;
     $object->{"lastcheck"}=0;
+    $object->{"lastreset"}=0;
+    $object->{"nextcheck"}=time();
     if ($object->{"method"} eq "tail") {
 	$object->reset_pointers;
     }
@@ -266,14 +272,14 @@ sub new {
 
 sub position {
     my $object=shift;
+    $object->{"endpos"}=sysseek($object->{handle},0,SEEK_END);
     unless ($object->{"tail"}) {
-	$object->logit("Start reading at end of file");
-	$object->{curpos}=sysseek($object->{handle},0,SEEK_END);
+	$object->{endpos}=$object->{curpos}=
+	    sysseek($object->{handle},0,SEEK_END);
     } elsif ($object->{"tail"}<0) {
-	$object->logit(" Start reading at start of file");
-	$object->{curpos}=sysseek($object->{handle},0,SEEK_START);
+	$object->{endpos}=sysseek($object->{handle},0,SEEK_END);
+	$object->{curpos}=sysseek($object->{handle},0,SEEK_SET);
     } else {
-	$object->logit(" Return ".$object->{"tail"}." lines."); 
 	my $crs=0;
 	my $maxlen=sysseek($object->{handle},0,SEEK_END);
 	while ($crs<$object->{"tail"}+1) {
@@ -285,9 +291,12 @@ sub position {
 	    $object->{curpos}=sysseek($object->{handle},-$calclen,SEEK_END);
 	    sysread($object->{handle},$object->{"buffer"},
 		    $calclen);
+	    $object->{curpos}=sysseek($object->{handle},0,SEEK_CUR);
 	    $crs=$object->{"buffer"}=~tr/\n//;
 	    last if ($calclen>=$maxlen);
 	}
+	$object->{curpos}=sysseek($object->{handle},0,SEEK_CUR);
+	$object->{endpos}=sysseek($object->{handle},0,SEEK_END);
 	if ($crs>$object->{"tail"}) {
 	    my $toskip=$crs-$object->{"tail"};
 	    my $pos;
@@ -303,7 +312,7 @@ sub position {
 
 sub reset_pointers {
     my $object=shift @_;
-    $object->logit("after ".(time()-$object->{lastread})."s\n");
+    $object->{lastreset} = time();
 
     my $st;
 
@@ -315,7 +324,7 @@ sub reset_pointers {
 	    return;
 	}
 	$object->error("Error opening ".$object->input.": $!");
-	return undef;
+	return;
     }
     
     if (defined($oldhandle)) {
@@ -331,26 +340,23 @@ sub reset_pointers {
         # last read, and replaced (within that second) with a file of equal
         # length, we're out of luck. I don't see how to fix this.
 	if (($st->mtime<int($object->{lastread}))) {
-	    $object->logit(" File not modified since last read. Reset skipped.");
-	    $object->logit("mtime:",$st->mtime," lastread: ",$object->{lastread});
+ 	    $object->{lastread} = $st->mtime; # repair fraction trouble
 	    return;
 	} elsif (($st->mtime==int($object->{lastread})) && 
 		 $st->size==$object->{"curpos"}) {
-	    $object->logit(" File not modified since last read. Reset skipped.");
-	    $object->logit("mtime:",$st->mtime," lastread: ",$object->{lastread});
+ 	    $object->{lastread} = $st->mtime; # repair fraction trouble
 	    return;
 
 	}
-	$object->logit("reseting file mtime:",$st->mtime," lastread: ",$object->{lastread}," size=",$st->size," read position=",$object->{"curpos"});
 	$object->{handle}=$newhandle;
 	$object->position;
+#	$object->{lastread} = $st->mtime; # repair fraction trouble
 	close($oldhandle);
     } else {                  # This is the first time we are opening this file
-	$object->logit(" First opening of file");
 	$st=stat($newhandle);
 	$object->{handle}=$newhandle;
 	$object->position;
-	$object->{lastread}=$st->mtime; # for better estimate on initial read
+#	$object->{lastread}=$st->mtime; # for better estimate on initial read
     }
     
 }
@@ -359,45 +365,37 @@ sub reset_pointers {
 sub checkpending {
    my $object=shift @_;
 
+   my $old_lastcheck = $object->{lastcheck};
    $object->{"lastcheck"}=time;
    unless ($object->{handle}) {
-       $object->logit(" Trying to find previously nonexistant file");
        $object->reset_pointers;
        unless ($object->{handle}) { # This try did not open the file either
-	   $object->logit(" File does not yet exist");
 	   return 0;
        }
    }
-   $object->logit("position = ".$object->{curpos}." interval = "
-	 .$object->interval);
    
-   my $endpos=sysseek($object->{handle},0,SEEK_END);
-   if ($endpos<$object->{curpos}) { 
-       $object->{curpos}=0;
-   } elsif (($object->{curpos}==$endpos) 
-	       && (time()-$object->{'lastread'})>$object->{'resetafter'}) {
+   $object->{"endpos"}=sysseek($object->{handle},0,SEEK_END);
+   if ($object->{"endpos"}<$object->{curpos}) {  # file was truncated
+       $object->{curpos}=sysseek($object->{handle},0,SEEK_SET);
+   } elsif (($object->{curpos}==$object->{"endpos"}) 
+	       && (time()-$object->{lastreset})>$object->{'resetafter'}) {
        $object->reset_pointers;
-       $endpos=sysseek($object->{handle},0,SEEK_END);
+       $object->{"endpos"}=sysseek($object->{handle},0,SEEK_END);
    }
 
-   if ($endpos-$object->{curpos}) {
+   if ($object->{"endpos"}-$object->{curpos}) {
        sysseek($object->{handle},$object->{curpos},SEEK_SET);
+       readin($object,$object->{"endpos"}-$object->{curpos});
    }
-   $object->logit("Returning length is ".($endpos-$object->{curpos}));
-   return ($endpos-$object->{curpos});
+   return ($object->{"endpos"}-$object->{curpos});
 }
 
 sub predict {
     my $object=shift;
     my $crs=$object->{"buffer"}=~tr/\n//; # Count newlines in buffer 
     my @call=caller(1);
-    $object->logit("called by ".join(" ",splice(@call,0,4)));
-    $object->logit("predict - $crs waiting in buffer");
     return 0 if $crs;
-    my $ttw=($object->{"lastcheck"}+$object->{"interval"})-time();
-    $object->logit("- lastcheck:".$object->{"lastcheck"}."interval="
-	  .$object->{"interval"});
-    $object->logit("- ttw is $ttw");
+    my $ttw=$object->{"nextcheck"}-time();
     return $ttw if $ttw>0;
     if (my $len=$object->checkpending) {
 	readin($object,$len);
@@ -408,8 +406,7 @@ sub predict {
 	$object->interval($object->interval*10);
     }
     $object->{"sleepcount"}++;
-    $object->logit("- checkpending returned 0 - (".$object->interval.",".
-	  $object->{"sleepcount"}.")");
+    $object->{"nextcheck"}=time()+$object->interval;
     return ($object->interval);
 }
 
@@ -419,7 +416,7 @@ sub bitprint {
 }
 
 sub select {
-    my $object=shift @_; # if ref($_[0]);
+    my $object=shift @_  if ref($_[0]);
     my ($timeout,@fds)=splice(@_,3);
     $object=$fds[0] unless defined($object);
     my ($savein,$saveout,$saveerr)=@_;
@@ -432,7 +429,6 @@ sub select {
     }
     foreach (@fds) {
 	my $val=$_->predict;
-	$object->logit("minpred calc 1 ($minpred,$val)");
 	$minpred=$val if $minpred>$val;
     }
     my ($nfound,$timeleft);
@@ -441,11 +437,9 @@ sub select {
 # Restore bitmaps in case we called select before
 	splice(@_,0,3,$savein,$saveout,$saveerr);
 
-	$object->logit("calling select(".bitprint($_[0]).",".bitprint($_[1]).",".bitprint($_[2]).",$minpred)\n");
 
 	($nfound,$timeleft)=select($_[0],$_[1],$_[2],$minpred);
 
-	$object->logit("returning ($nfound,$timeleft)=select(".bitprint($_[0]).",".bitprint($_[1]).",".bitprint($_[2]).",$minpred)\n");
 
 	if (defined($timeout)) {
 	    $minpred=$timeout;
@@ -456,9 +450,8 @@ sub select {
 	foreach (@fds) {
 	    my $val=$_->predict;
 	    $nfound++ unless $val;
-	    $object->logit("minpred calc 2 ($minpred,$val)");
 	    $minpred=$val if $minpred>$val;
-	    push(@retarr,$_);
+	    push(@retarr,$_) unless $val;
 	}
     }
     if (wantarray) {
@@ -476,18 +469,16 @@ sub readin {
 	$crs=$object->{"buffer"}=~tr/\n//; # Count newlines in buffer 
 	return $crs if $crs;
     }
+    $len=$object->{"maxbuf"} if ($len>$object->{"maxbuf"});
     my $nlen=$len;
     while ($nlen>0) {
 	$len=sysread($object->{handle},$object->{"buffer"},
 		     $nlen,length($object->{"buffer"}));
 	$nlen=$nlen-$len;
     }
-    $object->{curpos}=sysseek($object->{handle},
-			      $object->{curpos}+$len,SEEK_SET);
+    $object->{curpos}=sysseek($object->{handle},0,SEEK_CUR);
     
     $crs=$object->{"buffer"}=~tr/\n//;
-    $object->logit(" Got something ($len). there are now $crs newlines in buffer (".
-		   length($object->{"buffer"})." bytes)");
     
     if ($crs) {
 	my $tmp=time; 
@@ -500,20 +491,26 @@ sub readin {
 sub read {
     my $object=shift @_;
     my $len;
-    my $crs=$object->{"buffer"}=~tr/\n//; # Count newlines in buffer 
-    $object->logit("- $crs waiting in buffer");
-    while (!$crs) {
-	$object->logit(" Reading loop entered");
+    my $pending=$object->{"endpos"}-$object->{"curpos"};
+    my $crs=$object->{"buffer"}=~tr/\n//;
+    while (!$pending && !$crs) {
 	$object->{"sleepcount"}=0;
-	while (!($len=$object->checkpending) && !length($object->{"buffer"})) {
-	    return "" if $object->nowait;
-	    sleep($object->interval);               # maybe should be adjusted?
-	    if ($object->{"sleepcount"}++>$object->adjustafter) {
-		$object->{"sleepcount"}=0;
-		$object->interval($object->interval*10);
+	while ($object->predict) {
+	    if ($object->nowait) {
+		if (wantarray) {
+		    return ();
+		} else {
+		    return "";
+		}
 	    }
+	    sleep($object->interval);
 	}
-	next if $crs=readin($object,$len);
+	$pending=$object->{"endpos"}-$object->{"curpos"};
+	$crs=$object->{"buffer"}=~tr/\n//;
+    }
+    
+    if (!length($object->{"buffer"}) || index($object->{"buffer"},"\n")<0) {
+	readin($object,$pending);
     }
     unless (wantarray) {
 	my $str=substr($object->{"buffer"},0,
@@ -535,6 +532,7 @@ sub read {
 }
 
 1;
+
 __END__
 
 =head1 NAME
@@ -649,6 +647,17 @@ The number of seconds after last change when C<File::Tail> decides
 the file may have been closed and reopened. The default is 
 adjustafter*maxinterval.
 
+=item maxbuf
+
+The maximum size of the internal buffer. When File::Tail
+suddenly found an enormous ammount information in the file
+(for instance if the retry parameters were set to very
+infrequent checking and the file was rotated), File::Tail
+sometimes slurped way too much file into memory.  This sets
+the maximum size of File::Tail's buffer.
+
+Default value is 16384 (bytes).
+
 =item nowait
 
 Does not block on read, but returns an empty string if there is nothing
@@ -659,30 +668,38 @@ If you want to read tails from multiple files, use select.
 
 =item ignore_nonexistant
 
-    Do not complain if the file doesn't exist when it is first opened or
-when it is to be reopened. (File may be reopened after resetafter seconds 
-have passed since last data was found.)
+    Do not complain if the file doesn't exist when it is first 
+opened or when it is to be reopened. (File may be reopened after 
+resetafter seconds have passed since last data was found.)
 
 =item tail
-    When first started, read and return C<n> lines from the file. If C<n>
-is zero, start at the end of file. If C<n> is negative, return the whole file.
+
+    When first started, read and return C<n> lines from the file. 
+If C<n> is zero, start at the end of file. If C<n> is negative, 
+return the whole file.
+
     Default is C<0>.
 
 =item reset_tail
-    Same as tail, but applies after reset. (i.e. after the file has been
-automaticaly closed and reopened). Defaults to C<-1>, i.e. does not skip 
-any information present in the file when it first checks it. 
 
-   Why would you want it otherwise? I've seen files which have been cycled
-like this:
-   grep -v lastmonth log >newlog
-   mv log archive/lastmonth
-   mv newlog log
-   kill -HUP logger
-  
-   Obviously, if this happens and you have reset_tail set to c<-1>, you will
-suddenly get a whole bunch of lines - lines you already saw. So in this case, 
-reset_tail should probably be set to a small positive number or even C<0>.
+    Same as tail, but applies after reset. (i.e. after the
+file has been automaticaly closed and reopened). Defaults to
+C<-1>, i.e. does not skip any information present in the
+file when it first checks it.
+
+   Why would you want it otherwise? I've seen files which
+have been cycled like this:
+
+   grep -v lastmonth log >newlog 
+   mv log archive/lastmonth 
+   mv newlog log 
+   kill -HUP logger 
+
+
+Obviously, if this happens and you have reset_tail set to
+c<-1>, you will suddenly get a whole bunch of lines - lines
+you already saw. So in this case, reset_tail should probably
+be set to a small positive number or even C<0>.
 
 =item debug
 
@@ -762,9 +779,7 @@ Matija Grabnar, matija.grabnar@arnes.si
 =head1 SEE ALSO
 
 perl(1), tail (1), 
-MRTG 
-
-(http://ee-staff.ethz.ch/~oetiker/webtools/mrtg/mrtg.html)
+MRTG (http://ee-staff.ethz.ch/~oetiker/webtools/mrtg/mrtg.html)
 
 =cut
 
